from os.path import isfile
from glob import glob
from shlex import split
from subprocess import check_output


class GeoJSONConverter:
    def __init__(self, ogre_host='localhost:3000'):
        self.ogre_host = ogre_host

    def to_geojson(self, outfile, shapefile_prefix='', shapefile_ext='zip', shapefiles_dir='shapefiles'):

        if isfile(outfile):
            print "%s already exists. Skipping conversion..." % outfile
            return outfile

        print "Converting files in %s matching %s*.%s into GeoJSON. Results stored in %s" % \
              (shapefiles_dir, shapefile_prefix, shapefile_ext, outfile)

        with open(outfile, 'a') as out:
            file_pattern = "%s/%s*.%s" % (shapefiles_dir, shapefile_prefix, shapefile_ext)
            files = glob(file_pattern)

            for filename in files:
                try:
                    print "Converting %s to GeoJSON" % filename
                    raw_command = r'curl -F "upload=@%s" %s/convert' % (filename, self.ogre_host)
                    command = split(raw_command)
                    converted = check_output(command)
                    out.write("%s" % converted)
                except IOError as e:
                    print "Error converting %s" % filename
                    print "I/O error({0}): {1}".format(e.errno, e.strerror)



#!/usr/bin/python

import sys
from os.path import isdir
from os import mkdir
from getopt import getopt, GetoptError

from pyes import ES

from usshapes.mapper import Mapper
from usshapes.indexer import Indexer
from usshapes.builder import Builder
from usshapes.converter import GeoJSONConverter


class USShapesRunner():
    def __init__(self, args):
        es_host = 'localhost:9200'
        ogre_host = 'localhost:3000'
        batch_size = 100
        batch_mode = True
        excludes = []
        index_shapes = True
        index_suggestions = True

        try:
            opts, args = getopt(args, None, ["es-host=", "no-batch", "batch-size=", "excludes=", "skip-shapes", "skip-suggestions"])
        except GetoptError:
            print 'run.py [--es-host=<elasticsearch-host>] [[--no-batch] | [--batch-size=]] [--excludes=<excluded-types>] [--skip-shapes] [--skip-suggestions]'
            print """
                Options:
                - es-host: the elastic search host to use; default: localhost:9200
                - ogre-host: the ogre client to use; default: localhost:3000
                - no-batch: turn off batch mode
                - batch-size: how large should each batch be; default: 100
                - excludes: comma-separated list of excluded types; possible types: 'neighborhood', 'city', 'state', 'zip'
                - skip-shapes: skip creation and indexing of shapes
                - skip-suggestions: skip creation and indexing of suggestions
            """
            sys.exit(2)
        for opt, arg in opts:
            if opt == '--es-host':
                es_host = arg
            if opt == '--ogre-host':
                ogre_host = arg
            elif opt == '--no-batch':
                batch_mode = False
            elif opt == '--batch-size':
                batch_size = arg
            elif opt == '--excludes':
                excludes = [a.strip() for a in arg.split(',')]
            elif opt == '--skip-shapes':
                index_shapes = False
            elif opt == '--skip-suggestions':
                index_suggestions = False

        print "Running shapes and suggestions indexing with the following options:"
        print "* elasticsearch host: %s" % es_host
        print "* ogre host: %s" % ogre_host
        print "* batch mode: %s" % batch_mode
        print "* batch size: %s" % (batch_size if batch_mode else 'n/a')
        print "* excluded types: %s" % excludes
        print "* indexing shapes? %s" % index_shapes
        print "* indexing suggestions? %s" % index_suggestions

        es_client = ES(es_host)
        mapper = Mapper(es_client)
        indexer = Indexer(es_client, batch_size=batch_size, batch_mode=batch_mode)
        converter = GeoJSONConverter(ogre_host=ogre_host)
        builder = Builder(converter=converter)

        self.shapes_index = 'shapes_tmp'
        self.suggest_index = 'suggestions_tmp'

        self.initialize_mappings(mapper)

        if index_shapes:
            self.index_shapes(builder, indexer, excludes=excludes)
        if index_suggestions:
            self.index_suggestions(builder, indexer, excludes=excludes)

    def initialize_mappings(self, mapper):
        # Create indices and put mappings

        print 'Creating indices'
        mapper.create_indices(indices=[self.shapes_index, self.suggest_index])
        mapper.put_shapes_mappings(index=self.shapes_index)
        mapper.put_suggestions_mappings(index=self.suggest_index)

    def index_shapes(self, builder, indexer, excludes):
        # Download and format shapefiles, if they don't exist

        shapes_dir = 'shapes'
        neighborhood_shapes_file = '%s/shapes_neighborhood.json' % shapes_dir
        city_shapes_file = '%s/shapes_city.json' % shapes_dir
        state_shapes_file = '%s/shapes_state.json' % shapes_dir
        zip_shapes_file = '%s/shapes_zip.json' % shapes_dir

        if not isdir(shapes_dir):
            mkdir(shapes_dir)

        if 'neighborhood' not in excludes:
            builder.build_neighborhood_shapes(outfile=neighborhood_shapes_file)
            indexer.bulk_index(self.shapes_index, 'neighborhood', neighborhood_shapes_file)
        if 'city' not in excludes:
            builder.build_city_shapes(outfile=city_shapes_file)
            indexer.bulk_index(self.shapes_index, 'city', city_shapes_file)
        if 'state' not in excludes:
            builder.build_state_shapes(outfile=state_shapes_file)
            indexer.bulk_index(self.shapes_index, 'state', state_shapes_file)
        if 'zip' not in excludes:
            builder.build_zip_shapes(outfile=zip_shapes_file)
            indexer.bulk_index(self.shapes_index, 'zip', zip_shapes_file)

    def index_suggestions(self, builder, indexer, excludes):
        # Generate suggestion files, if they don't exist

        suggestions_dir = 'suggestions'
        neighborhood_suggestions_file = '%s/suggestions_neighborhood.json' % suggestions_dir
        city_suggestions_file = '%s/suggestions_city.json' % suggestions_dir
        zip_suggestions_file = '%s/suggestions_zip.json' % suggestions_dir

        shapes_dir = 'shapes'
        neighborhood_geofile = '%s/shapes_neighborhood.json' % shapes_dir
        city_geofile = '%s/shapes_city.json' % shapes_dir
        zip_geofile = '%s/shapes_zip.json' % shapes_dir

        if not isdir(suggestions_dir):
            mkdir(suggestions_dir)

        if 'neighborhood' not in excludes:
            builder.build_neighborhood_suggestions(outfile=neighborhood_suggestions_file,
                                                   neighborhood_geofile=neighborhood_geofile)
            indexer.bulk_index(self.suggest_index, 'neighborhood', neighborhood_suggestions_file)

        if 'city' not in excludes:
            builder.build_city_suggestions(outfile=city_suggestions_file, city_geofile=city_geofile)
            indexer.bulk_index(self.suggest_index, 'city', city_suggestions_file)

        if 'zip' not in excludes:
            builder.build_zip_suggestions(outfile=zip_suggestions_file, zip_geofile=zip_geofile)
            indexer.bulk_index(self.suggest_index, 'zip', zip_suggestions_file)


if __name__ == "__main__":
    USShapesRunner(sys.argv[1:])



import fileinput

from loaders import *
from utils import sanitize, state_codes


class Builder:
    good_line_re = re.compile(r'{\s*"type":\s*"Feature",\s*"properties":\s*')

    def __init__(self, converter):
        self.converter = converter
        self.raw_geo_dir = 'geojson'
        if not isdir(self.raw_geo_dir):
            mkdir(self.raw_geo_dir)

    def build_neighborhood_shapes(self, outfile):
        if isfile(outfile):
            print "%s already exists, skipping building raw geofile" % outfile
            return

        print "Building neighborhood shape files"

        shapefiles_dir = download_neighborhood_shapes()

        raw_geofile = '%s/raw_shapes_neighborhood.json' % self.raw_geo_dir
        self.converter.to_geojson(outfile=raw_geofile, shapefile_prefix='neighborhood', shapefiles_dir=shapefiles_dir)

        # Format results

        raw_geo_re = re.compile(
            '^(\{.*?)"STATE":\s*"([^"]+)".*?"CITY":\s*"([^"]+)".\s*"NAME":\s*"([^"]+)".*?\}.*"geometry":\s*(\{.*?\})\s*\},*$')

        doc_template = '{"id": "%(id)s", "state": "%(state)s", "city": "%(city)s", "neighborhood": "%(neighborhood)s", "geometry": %(coordinates)s}\n'

        with open(outfile, 'a') as out:
            print "Formatting %s into output file %s" % (raw_geofile, outfile)

            for line in fileinput.input(raw_geofile):
                is_good_line = self.good_line_re.match(line)
                if is_good_line is None:
                    continue

                m = raw_geo_re.match(line)

                (neighborhood, city, state, coordinates) = (m.group(4), m.group(3), m.group(2), m.group(5))
                id = sanitize("%s_%s_%s" % (neighborhood, city, state))

                data = {
                    'id': id,
                    'neighborhood': neighborhood,
                    'city': city,
                    'state': state,
                    'coordinates': coordinates
                }

                out.write(doc_template % data)

    def build_city_shapes(self, outfile):
        if isfile(outfile):
            print "%s already exists, skipping building geofile" % outfile
            return

        print "Building city shape files"

        shapefiles_dir = download_city_shapes()

        raw_geofile = '%s/raw_shapes_city.json' % self.raw_geo_dir
        self.converter.to_geojson(outfile=raw_geofile, shapefile_prefix='city', shapefiles_dir=shapefiles_dir)

        # Format results

        raw_geo_re = re.compile(
            '^\{.*?\{.*?\s*"STATEFP":\s*"([^"]+)".*?NAME":\s*"([^"]+)".*?\}.*"geometry":\s*(\{.*?\})\s*\},*$')

        doc_template = '{"id": "%(id)s", "state": "%(state)s", "city": "%(city)s", "geometry": %(coordinates)s}\n'

        with open(outfile, 'a') as out:
            print "Formatting %s into output file %s" % (raw_geofile, outfile)

            for line in fileinput.input(raw_geofile):
                is_good_line = self.good_line_re.match(line)
                if is_good_line is None:
                    continue

                m = raw_geo_re.match(line)

                (state_code, city, coordinates) = (m.group(1), m.group(2), m.group(3))
                state = state_codes[state_code] if state_code in state_codes else state_code
                id = sanitize("%s_%s" % (city, state))

                data = {
                    'id': id,
                    'city': city,
                    'state': state,
                    'coordinates': coordinates
                }

                out.write(doc_template % data)

    def build_state_shapes(self, outfile):
        if isfile(outfile):
            print "%s already exists, skipping building raw geofile" % outfile
            return

        print "Building state shape files"

        shapefiles_dir = download_state_shapes()

        raw_geofile = '%s/raw_shapes_state.json' % self.raw_geo_dir
        self.converter.to_geojson(outfile=raw_geofile, shapefile_prefix='state', shapefiles_dir=shapefiles_dir)

        # Format results

        raw_geo_re = re.compile(
            '^\{.*?: \{.*?"STUSPS": "([^"]+)", "NAME": "([^"]+)".*?\}.*"geometry":\s*(\{.*?\})\s*\},*$')

        doc_template = '{"id": "%(id)s", "state": "%(state)s", "postal": "%(postal)s", "geometry": %(coordinates)s}\n'

        with open(outfile, 'a') as out:
            print "Formatting %s into output file %s" % (raw_geofile, outfile)

            for line in fileinput.input(raw_geofile):
                is_good_line = self.good_line_re.match(line)
                if is_good_line is None:
                    continue

                m = raw_geo_re.match(line)

                (postal, state, coordinates) = (m.group(1), m.group(2), m.group(3))

                data = {
                    'id': sanitize(state),
                    'state': state,
                    'postal': postal,
                    'coordinates': coordinates
                }

                out.write(doc_template % data)

    def build_zip_shapes(self, outfile):
        if isfile(outfile):
            print "%s already exists, skipping building raw geofile" % outfile
            return

        print "Building zip shape files"

        shapefiles_dir = download_zip_shapes()

        raw_geofile = '%s/raw_shapes_zip.json' % self.raw_geo_dir
        self.converter.to_geojson(outfile=raw_geofile, shapefile_prefix='zip', shapefiles_dir=shapefiles_dir)

        # Format results

        raw_geo_re = re.compile('^\{.*?"ZCTA5CE10":\s*"([^"]+)".*?\}.*"geometry":\s*(\{.*?\})\s*\},*$')

        doc_template = '{"id": "%(id)s", "zipcode": "%(zipcode)s", "geometry": %(coordinates)s}\n'

        with open(outfile, 'a') as out:
            print "Formatting %s into output file %s" % (raw_geofile, outfile)

            for line in fileinput.input(raw_geofile):
                is_good_line = self.good_line_re.match(line)
                if is_good_line is None:
                    continue

                m = raw_geo_re.match(line)

                (zipcode, coordinates) = (m.group(1), m.group(2))

                data = {
                    'id': zipcode,
                    'zipcode': zipcode,
                    'coordinates': coordinates
                }

                out.write(doc_template % data)

    @staticmethod
    def build_neighborhood_suggestions(outfile, neighborhood_geofile):
        if isfile(outfile):
            print "%s already exists, skipping building raw geofile" % outfile
            return

        if not isfile(neighborhood_geofile):
            raise ValueError("a valid GeoJSON file must be supplied to build suggestions")

        print "Building neighborhood suggestion files"

        neighborhood_re = re.compile('^.*?"id": "([^"]+)", "state":\s*"([^"]+)",\s*"city":\s*"([^"]+)",\s*"neighborhood":\s*"([^"]+)".*')

        doc_template = '{"name" : "%(id)s","suggest" : { "input": "%(full_neighborhood)s", "output": "%(full_neighborhood)s", "payload" : {"type": "neighborhood", "id": "%(id)s"}}}\n'

        with open(outfile, 'a') as out:
            print "Formatting %s into output file %s" % (neighborhood_geofile, outfile)

            for line in fileinput.input(neighborhood_geofile):
                m = neighborhood_re.match(line)

                id = m.group(1)
                state = m.group(2).upper()
                city = m.group(3)
                neighborhood = m.group(4)

                full_neighborhood = ("%s, %s, %s" % (neighborhood, city, state))

                data = {
                    'id': id,
                    'full_neighborhood': full_neighborhood
                }

                out.write(doc_template % data)

    @staticmethod
    def build_city_suggestions(outfile, city_geofile):
        if isfile(outfile):
            print "%s already exists, skipping building raw geofile" % outfile
            return

        if not isfile(city_geofile):
            raise ValueError("a valid GeoJSON file must be supplied to build suggestions")

        print "Building city suggestion files"

        city_re = re.compile('^.*?"id": "([^"]+)", "state": "([^"]+)", "city": "([^"]+)".*')

        doc_template = '{"name" : "%(id)s","suggest" : {"input": "%(city)s, %(state)s","output": "%(city)s, %(state)s","payload": {"type": "city", "id": "%(id)s"}}}\n'

        with open(outfile, 'a') as out:
            print "Formatting %s into output file %s" % (city_geofile, outfile)

            for line in fileinput.input(city_geofile):
                m = city_re.match(line)

                id = m.group(1)
                state = m.group(2)
                city = m.group(3)

                data = {
                    'id': id,
                    'state': state,
                    'city': city
                }

                out.write(doc_template % data)


    @staticmethod
    def build_zip_suggestions(outfile, zip_geofile):
        if isfile(outfile):
            print "%s already exists, skipping building raw geofile" % outfile
            return

        if not isfile(zip_geofile):
            raise ValueError("a valid GeoJSON file must be supplied to build suggestions")

        print "Building zip suggestion files"

        city_re = re.compile('^.*?"id": "([^"]+)", "zipcode": "([^"]+)".*')

        doc_template = '{"name" : "%(id)s","suggest" : {"input": "%(zip)s", "output": "%(zip)s", "payload": {"type": "zip", "id": "%(id)s"}}}\n'

        with open(outfile, 'a') as out:
            print "Formatting %s into output file %s" % (zip_geofile, outfile)

            for line in fileinput.input(zip_geofile):
                m = city_re.match(line)

                id = m.group(1)
                zipcode = m.group(2)

                data = {
                    'id': id,
                    'zip': zipcode
                }

                out.write(doc_template % data)



from ftplib import FTP
from urllib2 import Request, urlopen
from os.path import isdir, isfile
from os import mkdir
import urllib
import re

from bs4 import BeautifulSoup
import requests


def download_state_shapes(outdir='shapefiles'):
    return download_shapefiles_from_census(type='STATE', outdir=outdir, outprefix='state')


def download_city_shapes(outdir='shapefiles'):
    return download_shapefiles_from_census(type='PLACE', outdir=outdir, outprefix='city')


def download_zip_shapes(outdir='shapefiles'):
    return download_shapefiles_from_census(type='ZCTA5', outdir=outdir, outprefix='zip')


# type should be one of PLACE, STATE, or ZCTA5
def download_shapefiles_from_census(type, outdir, outprefix):
    if not isdir(outdir):
        mkdir(outdir)

    census = FTP('ftp2.census.gov')
    print "Acquiring connection to census.gov"
    census.login()
    print "Connected!"
    filenames = census.nlst('geo/tiger/TIGER2013/%s' % type)

    for ftp_filename in filenames:
        zip_filename = re.match('.*/(.*\.zip)', ftp_filename).group(1)
        zip_filepath = '%s/%s%s' % (outdir, outprefix, zip_filename)

        if not isfile(zip_filepath):
            print 'Downloading %s' % zip_filename
            req = Request('ftp://ftp2.census.gov/%s' % ftp_filename)
            res = urlopen(req)
            with open(zip_filepath, 'a') as out:
                out.write(res.read())
        else:
            print '%s already exists. Skipping download' % zip_filepath

    census.close()

    return outdir


def download_neighborhood_shapes(outdir='shapefiles', outprefix='neighborhood'):
    zillow_neighborhoods_url = 'http://www.zillow.com/howto/api/neighborhood-boundaries.htm'
    result = requests.get(zillow_neighborhoods_url)
    soup = BeautifulSoup(result.text)
    links = [e.get('href') for e in soup.select('.illo-block ul a')]

    url_prefix = 'http://www.zillow.com'
    state_re = re.compile('-(.*?.zip)$')

    if not isdir(outdir):
        mkdir(outdir)

    f = urllib.URLopener()
    for link in links:
        state = state_re.search(link).group(1)
        filename = '%s/%s%s.zip' % (outdir, outprefix, state)

        if isfile(filename):
            print "%s already exists. Skipping..." % filename
            continue

        try:
            url = "%s%s" % (url_prefix, link)
            f.retrieve(url, filename)
        except IOError:
            print "Error retrieving zip file: %s" % url

    return outdir



from pyes.exceptions import IndexAlreadyExistsException


class Mapper:
    def __init__(self, client):
        self.client = client

    def create_indices(self, indices=[]):

        for index_name in indices:
            try:
                self.client.indices.create_index(index_name)
            except IndexAlreadyExistsException:
                print "[%s] index already exists. Skipping creation..." % index_name
                pass


    def put_shapes_mappings(self, index='shapes'):
        self.client.indices.put_mapping('neighborhood', {'properties': self.shapes_neighborhood_mapping}, [index])
        self.client.indices.put_mapping('city', {'properties': self.shapes_city_mapping}, [index])
        self.client.indices.put_mapping('state', {'properties': self.shapes_state_mapping}, [index])
        self.client.indices.put_mapping('zip', {'properties': self.shapes_zip_mapping}, [index])


    def put_suggestions_mappings(self, index='suggestions'):
        self.client.indices.put_mapping('neighborhood', {'properties': self.suggestions_mapping}, [index])
        self.client.indices.put_mapping('city', {'properties': self.suggestions_mapping}, [index])
        self.client.indices.put_mapping('zip', {'properties': self.suggestions_mapping}, [index])


    shapes_neighborhood_mapping = {
        "geometry": {
            "type": "geo_shape"
        },
        "city": {
            "type": "string"
        },
        "neighborhood": {
            "type": "string"
        },
        "state": {
            "type": "string"
        }
    }

    shapes_city_mapping = {
        "geometry": {
            "type": "geo_shape"
        },
        "city": {
            "type": "string"
        },
        "state": {
            "type": "string"
        }
    }

    shapes_state_mapping = {
        "geometry": {
            "type": "geo_shape"
        },
        "state": {
            "type": "string"
        },
        "postal": {
            "type": "string"
        }
    }

    shapes_zip_mapping = {
        "geometry": {
            "type": "geo_shape"
        },
        "zip": {
            "type": "string"
        }
    }

    suggestions_mapping = {
        "name": {
            "type": "string"
        },
        "suggest": {
            "type": "completion",
            "payloads": True
        }
    }


from fileinput import input
from time import sleep
from pyes.exceptions import NoServerAvailable
import re


class Indexer:
    def __init__(self, client, batch_mode=True, batch_size=100):
        self.client = client
        self.batch_mode = batch_mode
        self.client.bulk_size = int(batch_size)

    def bulk_index(self, index, type, shapefile, sleep_time=0.1):
        print 'Indexing [%s] docs into [%s] from %s' % (type, index, shapefile)

        index_count = 0

        id_re = re.compile('^.*?"id"\s*:\s*"([^"]+)"')
        parens_re = re.compile('\(.*?\)')

        for line in input(shapefile):
            id = id_re.match(line).group(1)

            # cleanup any lines that contain parentheticals
            line = parens_re.sub('', line).strip()

            # sweet dec/encodings bro
            line = line.decode('latin-1').encode('utf-8')
            id = id.decode('latin-1').encode('utf-8')

            try:
                self.client.index(line, index, type, id, bulk=self.batch_mode)
            except UnicodeDecodeError as e:
                print "Error processing line with id %s: %s" % (id, e.message)
            except NoServerAvailable as e:
                print "The server failed to respond while indexing %s: [%s]. Sleeping %d seconds and retrying..." % (id, e.message, sleep_time)
                sleep(5)
                try:
                    print "Retrying indexing of %s" % id
                    self.client.index(line, index, type, id, bulk=self.batch_mode)
                except NoServerAvailable as e:
                    print "Failed to reconnect again. Skipping indexing %s" % id
                except Exception as e:
                    print "This happened: %s" % e

            index_count += 1
            if index_count % int(self.client.bulk_size) == 0:
                print 'Indexing batch of %d, starting from %s' % (self.client.bulk_size, id)
                sleep(sleep_time)

        # index remaining bulk entries
        self.client.force_bulk()


#!/usr/bin/env python
from setuptools import setup

setup(name='usshapes',
      version='0.5.0',
      description='U.S. Shapefile Indexer for Elasticsearch',
      author='Daniel Sarfati',
      author_email='daniel@knockrentals.com',
      url='https://github.com/fatisar/us-shapes',
      scripts=['bin/us-shapes.py'],
      packages=['usshapes'],
      install_requires=['pyes', 'beautifulsoup4', 'requests']
)
